from base64 import b64decode
import json
import sys


# splice command types
CMD_TYPES = [4, 5, 6, 7, 255]


def i2b(i, wide):
    """
    i2b is a wrapper for int.to_bytes
    """
    return int.to_bytes(i, wide, byteorder="big")


def ifb(bites):
    """
    ifb is a wrapper for int.from_bytes
    """
    return int.from_bytes(bites, byteorder="big")


def parse_pid(byte1, byte2):
    """
    parse pid from packet
    """
    return (byte1 & 31) << 8 | byte2


def kv_clean(obj):
    """
    kv_clean removes items from a dict if the value is None
    """
    return {k: v for k, v in obj.items() if v is not None}


def kv_print(obj):
    """
    kv_print is a wrapper for printing
    a json dump of obj to sys.stderr
    """
    to_stderr(json.dumps(obj, indent=2))


def mk_bits(stuff):
    """
    Convert Hex and Base64 strings into bytes.
    """
    if stuff[:2].lower() == "0x":
        stuff = stuff[2:]
    if stuff[:2].lower() == "fc":
        return bytes.fromhex(stuff)
    try:
        return b64decode(stuff)
    except Exception:
        return stuff


def mk_payload(data):
    """
    mkpayload strips off packet headers
    when present
    """
    if data[0] == 0x47:
        payload = data[5:]
    else:
        payload = mk_bits(data)
    return payload


def get_json(obj):
    """
    get_json()
    returns obj as json.
    """
    return json.dumps(obj, indent=2)


def to_stderr(stuff):
    """
    Wrapper for printing to sys.stderr
    """
    print(stuff, file=sys.stderr)



import sys

from .cue import Cue
from .stream import Stream


def read_stdin():
    try:
        Stream(sys.stdin.buffer).decode()
    except:
        try:
            stuff = sys.stdin.buffer.read()
            Cue(stuff).show()
        except Exception:
            pass


def read_stuff(stuff):
    try:
        Cue(stuff).show()
    except Exception:
        try:
            with open(stuff, "rb") as tsdata:
                Stream(tsdata).decode()
        except Exception:
            pass


def decode(stuff=None):
    """
    All purpose SCTE 35 decoder function

    # for a mpegts video

        import threefive
        threefive.decode('/path/to/mpegts')

    # for a base64 encoded string

        import threefive
        Bee64='/DAvAAAAAAAA///wBQb+dGKQoAAZAhdDVUVJSAAAjn+fCAgAAAAALKChijUCAKnMZ1g='
        threefive.decode(Bee64)
    """
    if stuff in [None, sys.stdin.buffer]:
        read_stdin()
    else:
        read_stuff(stuff)



from bitn import BitBin
from .segmentation import SegmentationDescriptor
from .section import SpliceInfoSection
from .descriptors import (
    AvailDescriptor,
    DtmfDescriptor,
    TimeDescriptor,
    AudioDescriptor,
)
from .commands import (
    SpliceNull,
    SpliceSchedule,
    SpliceInsert,
    TimeSignal,
    BandwidthReservation,
    PrivateCommand,
)
from .tools import ifb, kv_clean, kv_print, mk_payload, to_stderr


class Cue:
    """
    The threefive.Splice class handles parsing
    SCTE 35 message strings.
    """

    # map of known descriptors and associated classes
    _descriptor_map = {
        0: AvailDescriptor,
        1: DtmfDescriptor,
        2: SegmentationDescriptor,
        3: TimeDescriptor,
        4: AudioDescriptor,
    }

    # map of known splice commands and associated classes
    _command_map = {
        0: SpliceNull,
        4: SpliceSchedule,
        5: SpliceInsert,
        6: TimeSignal,
        7: BandwidthReservation,
        255: PrivateCommand,
    }

    def __init__(self, data, packet_data=None):
        self.info_section = None
        self.command = None
        self.descriptors = []
        payload = mk_payload(data)
        self.packet_data = packet_data
        self._parse(payload)

    def _parse(self, payload):
        payload = self._mk_info_section(payload)
        payload = self._mk_command(payload)
        payload = self._mk_descriptors(payload)
        self.info_section.crc = hex(ifb(payload[0:4]))

    def _mk_info_section(self, payload):
        info_size = 14
        info_payload = payload[:info_size]
        self.info_section = SpliceInfoSection()
        self.info_section.decode(info_payload)
        return payload[info_size:]

    def _mk_command(self, payload):
        cmdbb = BitBin(payload)
        bit_start = cmdbb.idx
        self._set_splice_command(cmdbb)
        bit_end = cmdbb.idx
        cmdl = int((bit_start - bit_end) >> 3)
        self.command.splice_command_length = cmdl
        self.info_section.splice_command_length = cmdl
        return payload[cmdl:]

    def _mk_descriptors(self, payload):
        """
        parse descriptor loop length,
        then call Cue._descriptorloop
        """
        dll = ifb(payload[0:2])
        self.info_section.descriptor_loop_length = dll
        payload = payload[2:]
        self._descriptorloop(payload, dll)
        return payload[dll:]

    def __repr__(self):
        return str(self.get())

    def _descriptorloop(self, payload, dll):
        """
        parses all splice descriptors
        """
        while dll > 0:
            spliced = self._set_splice_descriptor(payload)
            sdl = spliced.descriptor_length
            bump = sdl + 2
            dll -= bump
            payload = payload[bump:]
            self.descriptors.append(spliced)

    def get(self):
        """
        Returns a dict of dicts for all three parts
        of a SCTE 35 message.
        """
        scte35 = {
            "info_section": self.get_info_section(),
            "command": self.get_command(),
            "descriptors": self.get_descriptors(),
        }
        if self.packet_data:
            scte35.update(self.get_packet_data())
        return scte35

    def get_command(self):
        """
        returns the SCTE 35
        splice command data as a dict.
        """
        return kv_clean(vars(self.command))

    def get_descriptors(self):
        """
        Returns a list of SCTE 35
        splice descriptors as dicts.
        """
        return [kv_clean(vars(d)) for d in self.descriptors]

    def get_info_section(self):
        """
        Returns SCTE 35
        splice info section as a dict
        """
        return kv_clean(vars(self.info_section))

    def get_packet_data(self):
        """
        returns cleaned Cue.packet_data
        """
        return kv_clean(self.packet_data)

    def _set_splice_command(self, cmdbb):
        """
        Splice Commands looked up in self._command_map
        """
        sct = self.info_section.splice_command_type
        if sct not in self._command_map.keys():
            to_stderr("Unknown Splice Command Type")
            return False
        self.command = self._command_map[sct]()
        self.command.decode(cmdbb)

    def _set_splice_descriptor(self, payload):
        """
        Splice Descriptors looked up in self._descriptor_map
        """
        # splice_descriptor_tag 8 uimsbf
        tag = payload[0]
        desc_len = payload[1]
        payload = payload[2:]
        bitbin = BitBin(payload[:desc_len])
        payload = payload[desc_len:]
        if tag in self._descriptor_map.keys():
            spliced = self._descriptor_map[tag](tag)
            spliced.decode(bitbin)
            spliced.descriptor_length = desc_len
            return spliced
        return False

    def show(self):
        """
        pretty prints the SCTE 35 message
        """
        kv_print(self.get())



import sys
from functools import partial

from bitn import BitBin
from .cue import Cue
from .streamtype import stream_type_map
from .tools import CMD_TYPES, parse_pid, to_stderr


def _parse_stream_type(bitbin):
    """
    extract stream pid and type
    """
    stream_type = bitbin.ashex(8)  # 8
    bitbin.forward(3)  # 11
    el_pid = bitbin.asint(13)  # 24
    bitbin.forward(4)  # 28
    eilib = bitbin.asint(12) << 3  # 40
    bitbin.forward(eilib)
    minus = 40 + eilib
    return minus, [stream_type, el_pid]


def _show_program_stream(pid, stream_type):
    """
    print program -> stream mappings
    """
    streaminfo = f"[{stream_type}] Reserved or Private"
    if stream_type in stream_type_map.keys():
        streaminfo = f"[{stream_type}] {stream_type_map[stream_type]}"
    to_stderr(f"\t   {pid}: {streaminfo}")


def show_cue(cue):
    """
    default function call for
    Stream.decode,
    Stream.decode_program,
    and Stream.decode_proxy
    when a SCTE-35 packet is found.
    """
    cue.show()


class Stream:
    """
    Stream class
    With MPEG-TS program awareness.
    Accurate pts for streams with
    more than one program containing
    SCTE-35 streams.
    """

    _CMD_TYPES = CMD_TYPES
    _PACKET_SIZE = 188

    def __init__(self, tsdata, show_null=False):
        self._tsdata = tsdata
        if show_null:
            self._CMD_TYPES.append(0)
        self._scte35_pids = set()
        self._pid_prog = {}
        self._pmt_pids = set()
        self._programs = set()
        self._prog_pts = {}
        self.info = False
        self.the_program = False

    def _find_start(self):
        """
        handles partial packets
        """
        pkt = self._tsdata.read(self._PACKET_SIZE)
        if pkt[0] == 71:
            return
        sync_byte = b"G"
        while self._tsdata:
            one_byte = self._tsdata.read(1)
            if not one_byte:
                sys.exit()
            if one_byte == sync_byte:
                self._tsdata.read(self._PACKET_SIZE - 1)
                if self._tsdata.read(1) is sync_byte:
                    self._tsdata.read(self._PACKET_SIZE - 1)
                    return

    def decode(self, func=show_cue):
        """
        reads MPEG-TS to find SCTE-35 packets
        """
        self._find_start()
        for pkt in iter(partial(self._tsdata.read, self._PACKET_SIZE), b""):
            cue = self._parser(pkt)
            if cue:
                func(cue)
            pkt = None

    def decode_next(self):
        """
        returns a threefive.Cue instance
        when a SCTE-35 packet is found
        """
        self._find_start()
        for pkt in iter(partial(self._tsdata.read, self._PACKET_SIZE), b""):
            cue = self._parser(pkt)
            if cue:
                return cue
            pkt = None

    def decode_program(self, the_program, func=show_cue):
        """
        returns a threefive.Cue instance
        when a SCTE-35 packet is found
        """
        self.the_program = the_program
        self.decode(func)

    def decode_proxy(self, func=show_cue):
        """
        reads an MPEG-TS stream
        and writes all ts packets to stdout
        and SCTE-35 data to stderr
        """
        self._find_start()
        for pkt in iter(partial(self._tsdata.read, self._PACKET_SIZE), b""):
            sys.stdout.buffer.write(pkt)
            cue = self._parser(pkt)
            if cue:
                func(cue)

    def show(self):
        """
        displays program stream mappings
        """
        self.info = True
        self.decode()

    def _mk_packet_data(self, pid):
        """
        creates packet_data dict
        to pass to a threefive.Cue instance
        """
        packet_data = {}
        packet_data["pid"] = pid
        prgm = self._pid_prog[pid]
        packet_data["program"] = prgm
        packet_data["pts"] = round(self._prog_pts[prgm], 6)
        return packet_data

    def _program_association_table(self, pkt):
        """
        parse program association table ( pid 0 )
        to program to program table pid mappings.
        """
        sectionlen = (pkt[6] & 15 << 8) | pkt[7]
        pkt = pkt[13 : (sectionlen + 5)]
        bitbin = BitBin(pkt)
        slib = sectionlen << 3
        slib -= 40
        while slib > 32:
            program_number = bitbin.asint(16)
            bitbin.forward(3)
            if program_number == 0:
                bitbin.forward(13)
            else:
                self._pmt_pids.add(bitbin.asint(13))
            slib -= 32
        # bitbin.forward(32)
        bitbin = None

    def _parser(self, pkt):
        """
        parse pid from pkt and
        route it appropriately
        """
        pid = parse_pid(pkt[1], pkt[2])
        if pid == 0:
            self._program_association_table(pkt)
            return
        if pid in self._pmt_pids:
            self._program_map_section(pkt)
            return
        if self.info:
            return
        if pid in self._scte35_pids:
            return self._parse_scte35(pkt, pid)
        if pid in self._pid_prog.keys():
            if (pkt[1] >> 6) & 1:
                pkt = pkt[0:18]
                self._parse_pusi(pkt, pid)
                return
        return

    def _parse_pts(self, pkt, pid):
        """
        parse pts
        """
        pts = ((pkt[13] >> 1) & 7) << 30
        pts |= ((pkt[14] << 7) | (pkt[15] >> 1)) << 15
        pts |= (pkt[16] << 7) | (pkt[17] >> 1)
        pts /= 90000.0
        ppp = self._pid_prog[pid]
        self._prog_pts[ppp] = pts

    def _parse_pusi(self, pkt, pid):
        """
        used to determine if pts data is available.
        """
        if pkt[6] & 1:
            if (pkt[10] >> 6) & 2:
                if (pkt[11] >> 6) & 2:
                    if (pkt[13] >> 4) & 2:
                        self._parse_pts(pkt, pid)

    def _parse_scte35(self, pkt, pid):
        """
        parse a SCTE-35 packet
        """
        packet_data = self._mk_packet_data(pid)
        # strip pre-padding on older scte-35 packets
        try:
            pkt = pkt[:5] + b"\xfc0" + pkt.split(b"\x00\xfc0")[1]
        except Exception:
            to_stderr("pre-strip failed")
        # check splice command type
        if pkt[18] in self._CMD_TYPES:
            return Cue(pkt, packet_data)
        return None

    def _parse_program_streams(self, slib, bitbin, program_number):
        """
        parse the elementary streams
        from a program
        """
        if program_number not in self._programs:
            self._programs.add(program_number)
            if self.info:
                to_stderr(f"\nProgram: {program_number}")
            while slib > 32:
                minus, pstream = _parse_stream_type(bitbin)
                slib -= minus
                stream_type = pstream[0]
                pid = pstream[1]
                self._pid_prog[pid] = program_number
                if self.info:
                    _show_program_stream(pid, stream_type)
                if stream_type == "0x86":
                    self._scte35_pids.add(pid)
        else:
            if self.info:
                sys.exit()

    def _program_map_section(self, pkt):
        """
        parse program maps for streams
        """
        # table_id = pkt[5]
        sectioninfolen = (pkt[6] & 15 << 8) | pkt[7]
        slib = sectioninfolen << 3
        program_number = (pkt[8] << 8) | pkt[9]
        # version = pkt[10] >> 1 & 31
        # current_next = pkt[10] & 1
        if self.the_program and (program_number != self.the_program):
            return None
        # section_number = pkt[11]
        # last_section_number = pkt[12]
        # pcr_pid = (pkt[13]& 31) << 8 | pkt[14]
        proginfolen = (pkt[15] & 15 << 8) | pkt[16]
        pkt = pkt[(17 + proginfolen) : (slib + 9)]
        bitbin = BitBin(pkt)
        pilib = proginfolen << 3
        slib -= 72
        slib -= pilib  # Skip descriptors
        self._parse_program_streams(slib, bitbin, program_number)
        bitbin = None
        return bitbin



stream_type_map = {
    "0x0": "ITU-T | ISO/IEC Reserved",
    "0x1": "ISO/IEC 11172 Video",
    "0x2": "ITU-T Rec. H.262 | ISO/IEC 13818-2 or ISO/IEC 11172-2 Video",
    "0x3": "ISO/IEC 11172 Audio",
    "0x4": "ISO/IEC 13818-3 Audio",
    "0x5": "ITU-T Rec. H.222.0 | ISO/IEC 13818-1 private sections",
    "0x6": "ITU-T Rec. H.222.0 | ISO/IEC 13818-1 PES packets- private data",
    "0x7": "ISO/IEC 13522 MHEG",
    "0x8": "ITU-T Rec. H.222.0 | ISO/IEC 13818-1 Annex A DSM-CC",
    "0x9": "ITU-T Rec. H.222.1",
    "0xA": "ISO/IEC 13818-6 type A",
    "0xB": "ISO/IEC 13818-6 type B",
    "0xC": "ISO/IEC 13818-6 type C",
    "0xD": "ISO/IEC 13818-6 type D",
    "0xE": "ITU-T Rec. H.222.0 | ISO/IEC 13818-1 auxiliary",
    "0xF": "ISO/IEC 13818-7 Audio with ADTS transport syntax",
    "0x10": "ISO/IEC 14496-2 Visual",
    "0x11": "ISO/IEC 14496-3 Audio - LATM transport ISO/IEC 14496-3 / AMD 1",
    "0x12": "ISO/IEC 14496-1 or FlexMux in PES packets",
    "0x13": "ISO/IEC 14496-1 or FlexMux in ISO/IEC14496 sections",
    "0x14": "ISO/IEC 13818-6 Synchronized Download Protocol",
    "0x1b": "Video",
    "0x86": "\033[92mSCTE 35\033[00m",
}



from .descriptors import SpliceDescriptor


"""
Table 20 from page 58 of
https://www.scte.org/SCTEDocs/Standards/ANSI_SCTE%2035%202019r1.pdf
"""
table20 = {
    0x00: "Restrict Group 0",
    0x01: "Restrict Group 1",
    0x02: "Restrict Group 2",
    0x03: "No Restrictions",
}

"""
table 22 from page 62 of
https://www.scte.org/SCTEDocs/Standards/ANSI_SCTE%2035%202019r1.pdf
I am using the segmentation_type_id as a key.

Segmentation_type_id : segmentation_message

"""
table22 = {
    0x00: "Not Indicated",
    0x01: "Content Identification",
    0x10: "Program Start",
    0x11: "Program End",
    0x12: "Program Early Termination",
    0x13: "Program Breakaway",
    0x14: "Program Resumption",
    0x15: "Program Runover Planned",
    0x16: "Program RunoverUnplanned",
    0x17: "Program Overlap Start",
    0x18: "Program Blackout Override",
    0x19: "Program Start ??? In Progress",
    0x20: "Chapter Start",
    0x21: "Chapter End",
    0x22: "Break Start",
    0x23: "Break End",
    0x24: "Opening Credit Start",
    0x25: "Opening Credit End",
    0x26: "Closing Credit Start",
    0x27: "Closing Credit End",
    0x30: "Provider Advertisement Start",
    0x31: "Provider Advertisement End",
    0x32: "Distributor Advertisement Start",
    0x33: "Distributor Advertisement End",
    0x34: "Provider Placement Opportunity Start",
    0x35: "Provider Placement Opportunity End",
    0x36: "Distributor Placement Opportunity Start",
    0x37: "Distributor Placement Opportunity End",
    0x38: "Provider Overlay Placement Opportunity Start",
    0x39: "Provider Overlay Placement Opportunity End",
    0x3A: "Distributor Overlay Placement Opportunity Start",
    0x3B: "Distributor Overlay Placement Opportunity End",
    0x40: "Unscheduled Event Start",
    0x41: "Unscheduled Event End",
    0x50: "Network Start",
    0x51: "Network End",
}


class SegmentationDescriptor(SpliceDescriptor):
    """
    Table 19 - segmentation_descriptor()
    """

    def __init__(self, tag):
        super().__init__(tag)
        self.name = "Segmentation Descriptor"
        self.segmentation_event_id = None
        self.segmentation_event_cancel_indicator = None
        self.component_count = None
        self.components = []
        self.program_segmentation_flag = None
        self.segmentation_duration_flag = None
        self.delivery_not_restricted_flag = None
        self.web_delivery_allowed_flag = None
        self.no_regional_blackout_flag = None
        self.archive_allowed_flag = None
        self.device_restrictions = None
        self.segmentation_duration = None
        self.segmentation_message = None
        self.segmentation_upid_type = None
        self.segmentation_upid_length = None
        self.segmentation_upid = None
        self.segmentation_type_id = None
        self.segment_num = None
        self.segments_expected = None
        self.sub_segment_num = None
        self.sub_segments_expected = None

    def decode(self, bitbin):
        self.parse_id(bitbin)
        self.segmentation_event_id = bitbin.ashex(32)  # 4 bytes
        self.segmentation_event_cancel_indicator = bitbin.asflag(1)
        bitbin.forward(7)  # 1 byte
        if not self.segmentation_event_cancel_indicator:
            self._set_flags(bitbin)  # 1 byte
            if not self.program_segmentation_flag:
                self._set_components(self, bitbin)
            self._set_segmentation(bitbin)

    def _set_components(self, bitbin):
        self.component_count = bitbin.asint(8)  # 1 byte
        for i in range(0, self.component_count):  # 6 bytes each
            comp = {}
            comp["component_tag"] = bitbin.asint(8)
            bitbin.forward(7)
            comp["pts_offset"] = bitbin.as90k(33)
            self.components[i] = comp

    def _set_flags(self, bitbin):  # 1 byte for set flags
        self.program_segmentation_flag = bitbin.asflag(1)
        self.segmentation_duration_flag = bitbin.asflag(1)
        self.delivery_not_restricted_flag = bitbin.asflag(1)
        if not self.delivery_not_restricted_flag:
            self.web_delivery_allowed_flag = bitbin.asflag(1)
            self.no_regional_blackout_flag = bitbin.asflag(1)
            self.archive_allowed_flag = bitbin.asflag(1)
            self.device_restrictions = table20[bitbin.asint(2)]
        else:
            bitbin.forward(5)

    def _set_segmentation(self, bitbin):
        if self.segmentation_duration_flag:
            self.segmentation_duration = bitbin.as90k(40)  # 5 bytes
        self.segmentation_upid_type = bitbin.asint(8)  # 1 byte
        self.segmentation_upid_length = bitbin.asint(8)  # 1 byte
        self.segmentation_upid = self._set_segmentation_upid(
            bitbin, self.segmentation_upid_type, self.segmentation_upid_length
        )
        self.segmentation_type_id = bitbin.asint(8)  # 1 byte
        if self.segmentation_type_id in table22.keys():
            self.segmentation_message = table22[self.segmentation_type_id]
            self._set_segments(bitbin)
        bitbin = None

    def _set_segmentation_upid(self, bitbin, upid_type, upid_length):
        upid_map = {
            0x02: ["Deprecated", self.AdID],
            0x03: ["Ad ID", self.AdID],
            0x04: ["UMID", self.UMID],
            0x05: ["ISAN", self.ISAN],
            0x06: ["ISAN", self.ISAN],
            0x07: ["TID", self.TID],
            0x08: ["AiringID", self.AirID],
            0x09: ["ADI", self.ADI],
            0x0A: ["EIDR", self.EIDR],
            0x0B: ["ATSC", self.ATSC],
            0x0C: ["MPU", self.MPU],
            0x0D: ["MID", self.MID],
            0x0E: ["ADS Info", self.ADS],
            0x0F: ["URI", self.URI],
        }
        upid_id = ""
        if upid_type in upid_map.keys():
            upid_id = upid_map[upid_type][1](bitbin, upid_length)
            if upid_type != 0x09:
                return f"{upid_map[upid_type][0]}:{upid_id}"
        return upid_id

    def _set_segments(self, bitbin):
        self.segment_num = bitbin.asint(8)  # 1 byte
        self.segments_expected = bitbin.asint(8)  # 1 byte
        if self.segmentation_type_id in [0x34, 0x36, 0x38, 0x3A]:
            """
            if there are 16 more bits in bitbin, read them.
            """
            if bitbin.idx >= 16:
                self.sub_segment_num = bitbin.asint(8)  # 1 byte
                self.sub_segments_expected = bitbin.asint(8)  # 1 byte
            else:
                self.sub_segment_num = self.sub_segments_expected = 0

    def ADI(self, bitbin, upid_length):
        return self.URI(bitbin, upid_length)

    def AdID(self, bitbin, upid_length):
        return self.URI(bitbin, upid_length)

    def ADS(self, bitbin, upid_length):
        return self.URI(bitbin, upid_length)

    def AirID(self, bitbin, upid_length):
        return bitbin.ashex(upid_length << 3)

    def ATSC(self, bitbin, upid_length):
        return {
            "TSID": bitbin.asint(16),
            "reserved": bitbin.asint(2),
            "end_of_day": bitbin.asint(5),
            "unique_for": bitbin.asint(9),
            "content_id": bitbin.asdecodedhex((upid_length - 4) << 3),
        }

    def ISAN(self, bitbin, upid_length):
        pre = "0000-0000-"
        middle = bitbin.ashex(upid_length << 3)
        post = "-0000-Z-0000-0000-6"
        return f"{pre}{middle[2:6]}{post}"

    def MID(self, bitbin, upid_length):
        upids = []
        bitcount = upid_length << 3
        while bitcount > 0:
            upid_type = bitbin.asint(8)  # 1 byte
            bitcount -= 8
            upid_length = bitbin.asint(8)
            bitcount -= 8
            segmentation_upid = self._set_segmentation_upid(
                bitbin, upid_type, upid_length
            )
            bitcount -= upid_length << 3
            upids.append(segmentation_upid)
        return upids

    def MPU(self, bitbin, upid_length):
        bitcount = upid_length << 3
        return {
            "format identifier": bitbin.asint(32),
            "private data": bitbin.asint(bitcount - 32),
        }

    def EIDR(self, bitbin, upid_length):
        pre = bitbin.asint(16)
        post = bitbin.ashex(80)
        return f"10.{pre}/{post[2:6]}-{post[6:10]}-{post[10:14]}-{post[14:18]}-T"

    def TID(self, bitbin, upid_length):
        return self.URI(bitbin, upid_length)

    def UMID(self, bitbin, upid_length):
        n = 8
        pre = "".join(self.AirID(bitbin, upid_length).split("x", 1))
        return ".".join([pre[i : i + n] for i in range(0, len(pre), n)])

    def URI(self, bitbin, upid_length):
        if upid_length > 0:
            return bitbin.asdecodedhex(upid_length << 3)
        return None



import sys
from .tools import to_stderr


class SpliceDescriptor:
    """
    SpliceDescriptor is the
    base class for all splice descriptors.
    It should not be used directly
    """

    def __init__(self, tag):
        self.tag = tag
        # identiﬁer 32 uimsbf == 0x43554549 (ASCII “CUEI”)
        self.identifier = None

    def decode(self, bitbin):
        """
        SpliceDescriptor subclasses
        must implement a decode method.
        """

    def encode(self):
        """
        SpliceDescriptor subclasses
        will soon implement a encode method.
        """

    def parse_id(self, bitbin):
        """
        parse splice descriptor identifier
        """
        self.identifier = bitbin.asdecodedhex(32)
        if self.identifier != "CUEI":
            to_stderr('Descriptors should have an identifier of "CUEI"')


class AvailDescriptor(SpliceDescriptor):
    """
    Table 17 -  avail_descriptor()
    """

    def __init__(self, tag):
        super().__init__(tag)
        self.name = "Avail Descriptor"
        self.provider_avail_id = None

    def decode(self, bitbin):
        self.parse_id(bitbin)
        self.provider_avail_id = bitbin.asint(32)


class DtmfDescriptor(SpliceDescriptor):
    """
    Table 18 -  DTMF_descriptor()
    """

    def __init__(self, tag):
        super().__init__(tag)
        self.name = "DTMF Descriptor"
        self.preroll = None
        self.dtmf_count = None
        self.dtmf_chars = []

    def decode(self, bitbin):
        self.parse_id(bitbin)
        self.preroll = bitbin.asint(8)
        self.dtmf_count = bitbin.asint(3)
        bitbin.forward(5)
        for i in range(0, self.dtmf_count):
            self.dtmf_chars[i] = bitbin.asint(8)


class TimeDescriptor(SpliceDescriptor):
    """
    Table 25 - time_descriptor()
    """

    def __init__(self, tag):
        super().__init__(tag)
        self.name = "Time Descriptor"
        self.tai_seconds = None
        self.tai_ns = None
        self.utc_offset = None

    def decode(self, bitbin):
        self.parse_id(bitbin)
        self.tai_seconds = bitbin.asint(48)
        self.tai_ns = bitbin.asint(32)
        self.utc_offset = bitbin.asint(16)


class AudioDescriptor(SpliceDescriptor):
    """
    Table 26 - audio_descriptor()
    """

    def __init__(self, tag):
        super().__init__(tag)
        self.name = "Audio Descriptor"
        self.components = []
        self.audio_count = None

    def decode(self, bitbin):
        self.parse_id(bitbin)
        self.audio_count = bitbin.asint(4)
        bitbin.forward(4)
        for i in range(0, self.audio_count):
            comp = {}
            comp["component_tag"] = bitbin.asint(8)
            comp["ISO_code="] = bitbin.asint(24)
            comp["bit_stream_mode"] = bitbin.asint(3)
            comp["num_channels"] = bitbin.asint(4)
            comp["full_srvc_audio"] = bitbin.asflag(1)
            self.components[i] = comp



#!/usr/bin/env python3

import setuptools
import threefive
with open("README.md", "r") as fh:
    readme = fh.read()

setuptools.setup(
    name="threefive",
    version=threefive.version() ,
    author="fu-corp",
    author_email="spam@futzu.com",
    description="threefive is The Leading SCTE-35 Parsing Lib for Mpeg-TS Video Streams, and Base64 or Hex Encoded Messages. threefive is Used in Production By Three of The Largest Broadcast Networks in The World..", 
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/futzu/SCTE35-threefive",
    packages=setuptools.find_packages(),
    install_requires=["bitn>=0.0.31",],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy"
    ],
    python_requires=">=3.6",
   
)



from base64 import b64encode
import sys
from bitn import BitBin
from threefive.tools import i2b, to_stderr


class SpliceInfoSection:
    """
    The SCTE-35 splice info section
    data.
    """

    def __init__(self):
        self.table_id = None
        self.section_syntax_indicator = None
        self.private = None
        self.reserved = None
        self.section_length = None
        self.protocol_version = None
        self.encrypted_packet = None
        self.encryption_algorithm = None
        self.pts_adjustment = None
        self.cw_index = None
        self.tier = None
        self.splice_command_length = None
        self.splice_command_type = None
        self.descriptor_loop_length = None

    def decode(self, bites):
        bitbin = BitBin(bites)
        self.table_id = bitbin.ashex(8)
        if self.table_id != "0xfc":
            to_stderr("splice info section table id should be 0xfc")
        self.section_syntax_indicator = bitbin.asflag(1)
        self.private = bitbin.asflag(1)
        self.reserved = bitbin.ashex(2)
        if self.reserved != "0x3":
            to_stderr("splice info section reserved should be 0x3")
        self.section_length = bitbin.asint(12)
        self.protocol_version = bitbin.asint(8)
        if self.protocol_version != 0:
            to_stderr("splice info section protocol version should be 0")
        self.encrypted_packet = bitbin.asflag(1)
        self.encryption_algorithm = bitbin.asint(6)
        self.pts_adjustment = bitbin.as90k(33)
        self.cw_index = bitbin.ashex(8)
        self.tier = bitbin.ashex(12)
        self.splice_command_length = bitbin.asint(12)
        self.splice_command_type = bitbin.asint(8)
        self.descriptor_loop_length = 0

    def encode(self):
        """
        first byte is:
            table_id
        """
        first_byte = int(self.table_id, 16)
        bencoded = i2b(first_byte, 1)
        """
        two_bytes is:
            section_syntax_indicator
            private
            reserved
            section_length
        """
        two_bytes = 0
        if self.section_syntax_indicator:
            two_bytes = 1 << 15
        if self.private:
            two_bytes += self.private << 14
        two_bytes += int(self.reserved, 16) << 12
        two_bytes += self.section_length
        bencoded += i2b(two_bytes, 2)
        """
        proto_byte is:
            protocol_version
        """
        proto_byte = self.protocol_version
        bencoded += i2b(proto_byte, 1)
        """
        five_bytes is:
            encrypted_packet
            encryption_algorithm
            pts_adjustment
        """
        five_bytes = 0
        if self.encrypted_packet:
            five_bytes = 1 << 39
        five_bytes += self.encryption_algorithm << 33
        five_bytes += int(self.pts_adjustment * 90000)
        bencoded += i2b(five_bytes, 5)
        """
        cw_byte is:
            cw_index
        """
        cw_byte = int(self.cw_index, 16)
        bencoded += i2b(cw_byte, 1)
        """
        three_bytes is:
            tier
            splice_command_length
        """
        three_bytes = int(self.tier, 16) << 12
        three_bytes += self.splice_command_length
        bencoded += i2b(three_bytes, 3)
        """
        cmd_byte is:
            splice_command_type
        """
        cmd_byte = self.splice_command_type
        bencoded += i2b(cmd_byte, 1)
        to_stderr(bencoded)
        to_stderr(b64encode(bencoded))



"""
threeefive.version

Used to set version in setup.py
and as an easy way to check which
version you have installed.
"""
version_tuple = ("2.2.27", "Two Two twenty-seven", "The 'Nobody puts baby in the corner' release")


def version():
    """
    version is set in the version_tuple
    to make it immutable.
    """
    return version_tuple[0]
